from io import BytesIO
from uuid import UUID

def _is_ascii(s):
    return all(ord(c) < 128 for c in s)


class IOBase:
    def __init__(self, file: str | BytesIO):
        self.file = file
    
    def get_pos(self):
        return self.file.tell()
    
    def set_pos(self, position: int, end=False):
        self.file.seek(-position if end else position, 2 if end else 0)
        
    def get_size(self):
        current_pos = self.file.tell()
        self.file.seek(0, 2)
        size = self.file.tell()
        self.file.seek(current_pos)
        return size
    
    def close(self):
        self.file.close()

class Reader(IOBase):
    def __init__(self, file=None):
        self.path = file
        if isinstance(file, str):
            self.file = open(file, 'rb')
        elif isinstance(file, bytes):
            self.file = BytesIO(file)
        else:
            self.file = BytesIO()
            
    def reopen(self):
        self.__init__(self.path)
    
    def read(self, size: int):
        return self.file.read(size)
        
    def uint(self):
        return int.from_bytes(self.file.read(1), byteorder='little')
    
    def uint32(self):
        return int.from_bytes(self.file.read(4), byteorder='little')
    
    def uint64(self):
        return int.from_bytes(self.file.read(8), byteorder='little')
    
    def int(self):
        return int.from_bytes(self.file.read(1), byteorder='little', signed=True)
    
    def int32(self):
        return int.from_bytes(self.file.read(4), byteorder='little', signed=True)
    
    def int64(self):
        return int.from_bytes(self.file.read(8), byteorder='little', signed=True)
    
    def sha1(self):
        return self.file.read(20)  
    
    def string(self, length=None):
        if length is None:
            length = self.int32()
        string = ""
        
        if length > 0:
            string = self.file.read(length).decode('ascii', errors='replace')
        elif length < 0:
            string = self.file.read(length * -2).decode('utf-16le', errors='replace')
        elif length == 0:
            string = ""
            
        return string.rstrip('\0')
    
    def strings_list(self, length=None):
        return self.list(self.string, length=length)
    
    def list(self, func, length=None):
        if not length:
            length = self.uint32()
        out_list = [func() for i in range(length)]
        return out_list
        
    def buffer(self, offset: int = None, size: int = None):
        if offset is None:
            offset = self.get_pos()
        
        current_pos = self.get_pos()
        self.set_pos(offset)
        buffer = Reader(self.read(size) if size else self.read())
        self.set_pos(current_pos)
        return buffer
        
class Writer(IOBase):
    def __init__(self, path = None):
        if path:
            self.file = open(path, "wb")
        else:
            self.file = BytesIO()
        
    def write(self, data):
        self.file.write(data)
        
    def uint(self, value: int):
        self.file.write(value.to_bytes(1, 'little'))
        
    def uint32(self, value: int):
        self.file.write(value.to_bytes(4, 'little'))
        
    def uint64(self, value: int):
        self.file.write(value.to_bytes(8, 'little'))
        
    def uint128(self, value: int):
        self.file.write(value.to_bytes(16, 'little'))
        
    def int(self, value: int):
        self.file.write(value.to_bytes(1, 'little', signed=True))

    def int32(self, value: int):
        self.file.write(value.to_bytes(4, 'little', signed=True))
        
    def int64(self, value: int):
        self.file.write(value.to_bytes(8, 'little', signed=True))
        
    def int128(self, value: int):
        self.file.write(value.to_bytes(16, 'little', signed=True))
        
        
    def sha1(self, value: bytes):
        self.file.write(value)  # Write 20 bytes for SHA1

    def string(self, value: str, use_unicode=False):
        value += '\x00'
        if (not use_unicode) and _is_ascii(value):
            self.uint32(len(value))
            self.write(value.encode("ascii"))
        else:
            length = int(len(value.encode("utf-16le")) / 2 * -1)
            self.int32(length)
            self.write(value.encode("utf-16le"))
            
    def list(self, list_items, use_length=False):
        if use_length:
            self.uint32(len(list_items))
        for item in list_items:
            self.write(item)
            
    def strings_list(self, list_items, use_length=False):
        if use_length:
            print("Length: ", len(list_items))
            self.uint32(len(list_items))
        for item in list_items:
            self.string(item)


from .file_io import Reader, Writer
from .version import PakVersion
from .entry import Entry
from .utils import fnv64, fnv64_path, split_path_child

import hashlib
from collections import defaultdict


def hash_sh1(data: bytes) -> bytes:
    hasher = hashlib.sha1()
    hasher.update(data)
    return hasher.digest()  # 20 bytes


class Index:
    def __init__(self):
        self.entry_count = 0

        self.entrys: dict[str, Entry] = {}


    def read(self, reader: Reader, version: PakVersion, index_offset: int, index_size: int):
        """Read the index of the pak file."""
        
        index_reader = reader.buffer(index_offset, index_size)
        
        self.mount_point = index_reader.string()
        self.entry_count = index_reader.uint32()
        path_hash_index = []
        directories = {}
        
        if version >= PakVersion.V10:
            path_hash_seed = index_reader.uint64()

            has_path_index = index_reader.uint32()
            if has_path_index != 0:
                path_hash_index_offset = index_reader.uint64()
                path_hash_index_size = index_reader.uint64()
                path_hash_index_hash = index_reader.sha1()
                
                path_hash_index_reader = reader.buffer(path_hash_index_offset, path_hash_index_size)
                count = path_hash_index_reader.uint32()
                for _ in range(count):
                    path_hash_index.append((
                        path_hash_index_reader.uint64(),  # Hash
                        path_hash_index_reader.uint32()   # encoded offset
                    ))
                
            has_full_directory_index = index_reader.uint32()
            if has_full_directory_index != 0:
                full_directory_index_offset = index_reader.uint64()
                full_directory_index_size = index_reader.uint64()
                full_directory_index_hash = index_reader.sha1()
                
                full_directory_index_reader = reader.buffer(full_directory_index_offset, full_directory_index_size)
                dir_count = full_directory_index_reader.uint32()
                for _ in range(dir_count):
                    dir_name = full_directory_index_reader.string()
                    file_count = full_directory_index_reader.uint32()
                    files = {}
                    for _ in range(file_count):
                        file_name = full_directory_index_reader.string()
                        files[file_name] = full_directory_index_reader.int32()
                    directories[dir_name] = files
                    
            size = index_reader.int32()
            encoded_entries = index_reader.read(size)
            
            not_encoded_entry_count = index_reader.uint32()
            not_encoded_entries = []
            for _ in range(not_encoded_entry_count):
                entry = Entry()
                entry.read(index_reader, version)
                self.entrys.append(entry)
                
            entrys_by_path = {}
            if has_full_directory_index:
                encoded_offset_reader = Reader(encoded_entries)
                for dir_name, files in directories.items():
                    for file_name, encoded_offset in files.items():
                        entry = None
                        if encoded_offset >= 0:
                            encoded_offset_reader.set_pos(encoded_offset)
                            entry = Entry()
                            entry.read_encoded(encoded_offset_reader, version)
                        else:
                            index = -encoded_offset - 1
                            entry = not_encoded_entries[index].copy()
                        
                        path = dir_name.lstrip("/") + file_name
                        entrys_by_path[path] = entry
                        
            self.entrys = entrys_by_path

        else:
            self.entrys = {}
            for _ in range(self.entry_count):
                entry = Entry()
                path = index_reader.string()
                entry.read(index_reader, version)
                self.entrys[path] = entry
        
        
    def write(self, writer: Writer, version: PakVersion, mount_point: str = "../../../", 
              path_hash_seed: int = 0):
        self.offset = writer.get_pos()
        
        index_buf = Writer()
        
        fdi_buf, phi_buf = None, None

        index_buf.string(mount_point) # mount point
        index_buf.uint32(len(self.entrys)) # entry count
        
        if version >= PakVersion.V10:
            index_buf.uint64(path_hash_seed) # path_hash_seed

            encoded_entries = Writer()
            entrys_offsets = []
            for path, entry in self.entrys.items():
                entrys_offsets.append((path, encoded_entries.get_pos()))
                entry.write_encoded(encoded_entries, version)
                
            bytes_before_phi = (
                105 + len(mount_point) + len(encoded_entries.file.getvalue())
            )
            
            phi_offset = self.offset + bytes_before_phi
            phi_buf = Writer()
            
            generate_phi(phi_buf, entrys_offsets, path_hash_seed)
            
            fdi_offset = phi_offset + len(phi_buf.file.getvalue())
            fdi_buf = Writer()
            
            generate_fdi(fdi_buf, entrys_offsets)
            
            index_buf.uint32(1)
            index_buf.uint64(phi_offset)
            index_buf.uint64(len(phi_buf.file.getvalue()))
            index_buf.write(hash_sh1(phi_buf.file.getvalue()))
            
            index_buf.uint32(1)
            index_buf.uint64(fdi_offset)
            index_buf.uint64(len(fdi_buf.file.getvalue()))
            index_buf.write(hash_sh1(fdi_buf.file.getvalue()))
            
            index_buf.uint32(len(encoded_entries.file.getvalue()))
            index_buf.write(encoded_entries.file.getvalue())
            index_buf.uint32(0)
        else:
            for path, entry in self.entrys.items():
                index_buf.string(path)
                entry.write(index_buf, version)
                
        self.hash = hash_sh1(index_buf.file.getvalue())
        self.size = len(index_buf.file.getvalue())
        
        writer.write(index_buf.file.getvalue())
        
        if phi_buf:
            writer.write(phi_buf.file.getvalue())
            
        if fdi_buf:
            writer.write(fdi_buf.file.getvalue())


def generate_phi(writer: Writer, entries: list[tuple[str, int]], path_hash_seed=0):
    writer.uint32(len(entries))
    for path, offset in entries:
        path_hash = fnv64_path(path, path_hash_seed)
        writer.uint64(path_hash)
        writer.uint32(offset)
        
    writer.uint32(0)
   
    
def generate_fdi(writer: Writer, entries: list[tuple[str, int]]):
    fdi = defaultdict(dict)

    for path, offset in entries:
        p = path
        while True:
            result = split_path_child(p)
            if result is None:
                break
            parent, _ = result
            p = parent
            fdi.setdefault(parent, {})

        result = split_path_child(path)
        if result is None:
            raise ValueError("Path is root — invalid in this context")
        dir, filename = result
        fdi[dir][filename] = offset
        
    writer.uint32(len(fdi))
    for dir, files in fdi.items():
        writer.string(dir)
        writer.uint32(len(files))
        for filename, offset in files.items():
            writer.string(filename)
            writer.uint32(offset)


from .file_io import Reader, Writer
from .version import PakVersion


PAK_MAGIC = 0x5A6F12E1


def check_pak_version(reader: Reader) -> PakVersion:
    """Check the version of the pak file."""
    
    size = reader.get_size()
    
    reader.set_pos(size - 44) # Version from 1 to 7
    magic = reader.uint32()
    if magic == PAK_MAGIC:
        return PakVersion(reader.uint32())
    
    reader.set_pos(size - 172) # Version 8A
    magic = reader.uint32()
    if magic == PAK_MAGIC:
        return PakVersion.V8A
    
    reader.set_pos(size - 204) # Version 8B, 10, 11
    magic = reader.uint32()
    if magic == PAK_MAGIC:
        return PakVersion(reader.uint32() + 1)   
    
    reader.set_pos(size - 205) # Version 9
    magic = reader.uint32()
    if magic == PAK_MAGIC:
        return PakVersion.V9
    
    raise ValueError("Invalid pak file or unsupported version.")


class Footer():
    def __init__(self):
        self.encryption_key = None
        self.is_encrypted = False
        self.version = None
        self.index_offset = 0
        self.index_size = 0
        self.hash = None
        self.is_frozen = False
        self.compresion = None
        
    def read(self, reader: Reader):
        """Read the footer of the pak file."""
        
        self.version = check_pak_version(reader)
        if self.version < PakVersion.V4:
            reader.set_pos(44, end=True)
        elif self.version < PakVersion.V7:
            reader.set_pos(45, end=True)
        elif self.version == PakVersion.V7:
            reader.set_pos(65, end=True)
        elif self.version == PakVersion.V8A:
            reader.set_pos(193, end=True)
        elif self.version == PakVersion.V9:
            reader.set_pos(226, end=True)
        else:
            reader.set_pos(225, end=True)
        
        if self.version >= PakVersion.V7:
            self.encryption_key = reader.sha1() 
        if self.version >= PakVersion.V4:
            self.is_encrypted = reader.uint() == 1
        
        magic = reader.uint32()
        version = reader.uint32()
        
        self.index_offset = reader.uint64()
        self.index_size = reader.uint64()
        print(f"Footer: magic={magic}, version={version}, index_offset={self.index_offset}, index_size={self.index_size}")
        self.hash = reader.sha1()
        self.is_frozen = reader.uint() == 1 if self.version >= PakVersion.V9 else False
            
        # TODO: COMPRESSION
    
    
    def write(self, writer: Writer, version: PakVersion, offset, size, hash):
        if version >= PakVersion.V7:
            writer.uint128(0) 
        if version >= PakVersion.V4:
            writer.uint(0) # is encrypted TODO
            
        writer.uint32(PAK_MAGIC)
        
        if version >= 9:
            writer.uint32(version - 1)
        else:
            writer.uint32(version)
            
        writer.uint64(offset)
        writer.uint64(size)
        writer.write(hash)
        
        if version == PakVersion.V9:
            writer.uint(0) # is frozen
        
        algo_size = 0
        if version == PakVersion.V8A:
            algo_size = 4
        elif version > PakVersion.V8A:
            algo_size = 5
                        
        writer.write(b"\x00" * 32 * algo_size)





def hybrid_method(func):
    def wrapper(instance_or_cls, *args, **kwargs):
        if isinstance(instance_or_cls, type):
            # Called on class: create a temporary instance
            temp_instance = instance_or_cls()
            return func(temp_instance, *args, **kwargs)
        else:
            # Called on instance
            return func(instance_or_cls, *args, **kwargs)

    return classmethod(wrapper)

        
        
def fnv64(data_bytes: bytes, offset: int) -> int:
    OFFSET = 0xcbf29ce484222325
    PRIME = 0x00000100000001b3
    hash_ = (OFFSET + offset) & 0xFFFFFFFFFFFFFFFF

    for b in data_bytes:
        hash_ ^= b
        hash_ = (hash_ * PRIME) & 0xFFFFFFFFFFFFFFFF  # simulate u64 wrapping

    return hash_


def fnv64_path(path: str, offset: int) -> int:
    lower = path.lower()
    utf16le = lower.encode("utf-16le")  # match encode_utf16 + to_le_bytes
    return fnv64(utf16le, offset)


def split_path_child(path: str) -> tuple[str, str] | None:
    if path == "/" or not path:
        return None

    # Remove trailing slash if present
    path = path.rstrip("/")

    idx = path.rfind("/")
    if idx != -1:
        return (path[:idx + 1], path[idx + 1:])
    else:
        return ("/", path)


from .file_io import Reader, Writer
from .version import PakVersion
from .utils import hybrid_method

import hashlib
from math import ceil

from cryptography.hazmat.primitives.ciphers import Cipher, algorithms, modes
from cryptography.hazmat.backends import default_backend


UINT32_MAX = 0xFFFFFFFF


def align(offset: int) -> int:
    return (offset + 15) & ~15


class Block:
    """Represents a block in the entry."""
    
    def __init__(self, start, end):
        self.start = start
        self.end = end
        
    def read(self, reader: Reader):
        """Read the block from the reader."""
        self.start = reader.uint64()
        self.end = reader.uint64()
        
        return self


class Entry:
    def __init__(self):
        self.offset = 0
        self.is_encrypted = False
        self.data = None
        self.blocks: list[Block] = []
        self.compressio_name = None
        self.compression_block_size = 0
        self.compression_block_count = 0
        self.size = 0
        self.compressed_size = 0
        
    def __repr__(self):
        return f"Entry(offset={self.offset}, size={self.size})"


    @staticmethod
    def get_serialized_size(version: PakVersion, size: int, compression_block_count: int) -> int:
        out = 24
        if version == PakVersion.V8A:
            out += 1
        else:
            out += 4
        
        if version == PakVersion.V1:
            out += 8
        
        out += 20
        
        if size:
            out += 4 + 16 * compression_block_count
            
        size += 1
        
        if version >= PakVersion.V3:
            out += 4
        
        return out


    @hybrid_method
    def read(self, reader:Reader, version: PakVersion) -> 'Entry':
        """Read the entry from the reader."""
        self.offset = reader.uint64()
        self.compressed_size = reader.uint64()
        self.size = reader.uint64()
        
        if version == PakVersion.V8A:
            self.compressio_name = reader.uint()
        else:
            self.compressio_name = reader.uint32()
        self.compressio_name = None if self.compressio_name == 0 else self.compressio_name - 1
            
        if version == PakVersion.V1:
            self.timestamp = reader.uint64()
            
        self.hash = reader.sha1()
            
        if version >= PakVersion.V3:
            if self.compressio_name is not None:
                self.blocks = reader.list(Block().read)
            self.is_encrypted = reader.uint()
            self.compressio_block_size = reader.uint32()
            
        return self
            
    @hybrid_method
    def read_encoded(self, reader: Reader, version: PakVersion):
        """Read the entry from the encoded reader."""
        
        data = reader.uint32()
        self.compressio_name = (data >> 23) & 0x3F
        self.compressio_name = None if self.compressio_name == 0 else self.compressio_name - 1
        
        self.is_encrypted = bool(data & (1 << 22))
        compression_block_count = (data >> 6) & 0xffff
        compression_block_size = data & 0x3f
        
        if compression_block_size == 0x3F:
            compression_block_size = reader.uint32()
        else:
            compression_block_size <<= 11
            
            
        def read_varint(bit):
            if data & (1 << bit) != 0 :
                return reader.uint32()
            else:
                return reader.uint64()
            
        self.offset = read_varint(31)
        self.size = read_varint(30)

        if self.compressio_name is not None:
            self.compressed_size = read_varint(29)
        else:
            self.compressed_size = self.size
            
        offset_base = Entry.get_serialized_size(version, self.size, compression_block_count)
        
        if compression_block_count == 1 and not self.is_encrypted:
            self.blocks = [Block(offset_base, offset_base + self.size)]
        elif compression_block_count > 0:
            index = offset_base
            self.blocks = []
            for _ in range(compression_block_count):
                block_size = reader.uint32()
                block = Block(index, index + block_size)
                if self.is_encrypted:
                    block_size = align(block_size)
            
                index += block_size
                self.blocks.append(block)
            
        return self
    
    @hybrid_method
    def read_file(self, reader: Reader, version: PakVersion, key: bytes):
        """Read the entry from the file."""
        reader.set_pos(self.offset)

        test = Entry.read(reader, version)
        
        data = None
        if self.is_encrypted:
            data = reader.read(align(self.compressed_size))
        else:
            data = reader.read(self.size)
            
        if self.is_encrypted:
            cipher = Cipher(algorithms.AES(key), modes.ECB(), backend=default_backend())
            decryptor = cipher.decryptor()
            
            decrypted_data = bytearray()
            for i in range(0, len(data), 16):
                block = data[i:i+16]
                if len(block) == 16:  # Only decrypt full blocks
                    decrypted_block = decryptor.update(block)
                    decrypted_data.extend(decrypted_block)
                else:
                    decrypted_data.extend(block)
            
            # Truncate to original compressed size
            data = decrypted_data[:self.compressed_size]
    
        return bytes(data)
    
    def write_data(self, writer: Writer, version: PakVersion):
        self.offset = writer.get_pos()
        self.size = len(self.data)
        self.compressed_size = self.size
        
        writer.uint64(self.offset) # offset
        writer.uint64(self.size) # size
        writer.uint64(self.compressed_size) # uncomp size 
        
        # TODO: Compression
        
        if version == PakVersion.V8A:
            writer.uint(0)
        else:
            writer.uint32(0)
            
        if version == PakVersion.V1:
            writer.uint64(0)
                
        hash = hashlib.sha1()
        hash.update(self.data)
        self.hash = hash.digest()
        writer.sha1(self.hash)
        
        if version >= PakVersion.V3:
            if self.compressio_name is not None:
                pass # TODO: COMPRESSION
            self.flags = writer.uint(0) # TODO
            self.compressio_block_size = writer.uint32(0) # 0 if not compressed
        
        writer.write(self.data)

    
    def write(self, writer: Writer, version: PakVersion):
        writer.uint64(self.offset) # offset
        writer.uint64(self.size) # size
        writer.uint64(self.compressed_size) # uncomp size
        # TODO: COMP SUP
        
        # COMP NAME 
        # TODO: COMP
        if version == PakVersion.V8A:
            writer.uint(0)
        else:
            writer.uint32(0)
            
        if version == PakVersion.V1:
            writer.uint64(0)
                
        writer.sha1(self.hash)
        
        if version >= PakVersion.V3:
            if self.compressio_name is not None:
                pass # TODO: Compression
            self.flags = writer.uint(0) # TODO
            self.compressio_block_size = writer.uint32(0) # 0 if not compressed
    
    
    def write_encoded(self, writer: Writer, version: PakVersion):
        compression_block_size = (self.compression_block_size >> 11) & 0x3F
        if (compression_block_size << 11) != self.compression_block_size:
            self.compression_block_size = 0x3F
        
        is_size_32_bit_safe = self.compressed_size <= UINT32_MAX
        is_uncompressed_size_32_bit_safe = self.size <= UINT32_MAX
        is_offset_32_bit_safe = self.offset <= UINT32_MAX
        
        flags = (
            compression_block_size
            | (self.compression_block_count << 6)
            | (int(self.is_encrypted) << 22)
            | ((self.compressio_name + 1 if self.compressio_name is not None else 0) << 23)
            | (int(is_size_32_bit_safe) << 29)
            | (int(is_uncompressed_size_32_bit_safe) << 30)
            | (int(is_offset_32_bit_safe) << 31)
        )
        
        writer.uint32(flags)
        
        if self.compression_block_size == 0x3F:
            writer.uint32(self.compression_block_size)
            
        if is_offset_32_bit_safe:
            writer.uint32(self.offset)
        else:
            writer.uint64(self.offset)
            
        if is_uncompressed_size_32_bit_safe:
            writer.uint32(self.size)
        else:
            writer.uint64(self.size)


from .version import PakVersion
from .file_io import Reader, Writer
from .footer import Footer
from .index import Index
from .entry import Entry

import shutil, os
class PakFile:
    def __init__(self):
        """Initialize a new PakFile instance."""
        self.version = PakVersion.V11
        self.key = bytes(32)
        self.mount_point = "../../../"
        self.path_hash_seed = 0
        
        self.reader = Reader()
        
        self._index = Index()
        self._footer = Footer()
        
        
    @property
    def count(self):
        """Get the number of entries in the pak file."""
        return len(self._index.entrys)
    
    
    def set_key(self, key: bytes | str):
        """Set the encryption key for the pak file."""
        byte_key = bytes.fromhex(key) if isinstance(key, str) else key

        if len(byte_key) != 32:
            raise ValueError("Key must be 32 bytes long.")
        
        self.key = byte_key
        
    def set_mount_point(self, mount_point: str):
        """Set the mount point for the pak file."""
        self.mount_point = mount_point
        
    def set_path_hash_seed(self, seed: int):
        """Set the path hash seed for the pak file."""
        if not isinstance(seed, int):
            raise ValueError("Seed must be an integer.")
        
        self.path_hash_seed = seed
        
    def set_version(self, version: PakVersion):
        """Set the version of the pak file."""
        if not isinstance(version, PakVersion):
            raise ValueError("Version must be an instance of PakVersion.")
        
        self.version = version


    def read(self, file: str | bytes):
        """Read the pak file."""
        
        self.reader = Reader(file)
        
        self._footer = Footer()
        self._footer.read(self.reader)
        
        self.version = self._footer.version
        
        self._index = Index()
        self._index.read(self.reader, 
            self._footer.version,
            self._footer.index_offset,
            self._footer.index_size)
        
        
    def write(self, file: str):
        """Write the pak file."""
        
        self.reader.reopen()
        writer = Writer(f"{file}.tmp")
        
        for _, entry in self._index.entrys.items():
            if entry.data:
                entry.write_data(writer, self.version)
            else:
                entry.data = entry.read_file(self.reader, self.version, self.key)
                entry.write_data(writer, self.version)
                
            entry.data = None
            
        self.reader.close()
            
        self._index.write(writer, self.version)
        self._footer.write(writer, self.version, self._index.offset, self._index.size, self._index.hash)
        writer.close()
        
        shutil.move(f"{file}.tmp", file)


    def add_file(self, path: str, data: bytes):
        """Add a file to the pak file."""
        entry = Entry()
        self._index.entrys[path] = entry
        entry.data = data
        
        
    def remove_file(self, path: str):
        """Remove a file from the pak file."""

        if path in self._index.entrys:
            del self._index.entrys[path]
        else:
            raise KeyError(f"Path '{path}' not found in pak file.")
                
        
    def read_file(self, path: str) -> bytes:
        """Read a file from the pak file."""
        
        entry = self._index.entrys.get(path)
        return entry.read_file(self.reader, self._footer.version, self.key) 
    
    
    def list_files(self) -> list[str]:
        """List all files in the pak file."""
        return list(self._index.entrys.keys())
    
    
    def unpack(self, output_dir: str):
        """Unpack the pak file to the specified directory."""

        if not os.path.exists(output_dir):
            os.makedirs(output_dir)

        entries_list = self.list_files()

        for path in entries_list:
            entry = self._index.entrys.get(path)
            file_path = os.path.join(output_dir, path)
            os.makedirs(os.path.dirname(file_path), exist_ok=True)

            data = entry.read_file(self.reader, self._footer.version, self.key)
            with open(file_path, 'wb') as f:
                f.write(data)
